#!/usr/bin/env python
"""
This module provides a class for cron-like scheduling systems, and
exposes the function used to convert static cron expressions to Python
sets.

CronExpression objects are instantiated with a cron formatted string
that represents the times when the trigger is active. When using
expressions that contain periodic terms, an extension of cron created
for this module, a starting epoch should be explicitly defined. When the
epoch is not explicitly defined, it defaults to the Unix epoch. Periodic
terms provide a method of recurring triggers based on arbitrary time
periods.


Standard Cron Triggers:
>>> job = CronExpression("0 0 * * 1-5/2 find /var/log -delete")
>>> job.check_trigger((2010, 11, 17, 0, 0))
True
>>> job.check_trigger((2012, 12, 21, 0 , 0))
False

Periodic Trigger:
>>> job = CronExpression("0 %9 * * * Feed 'it'", (2010, 5, 1, 7, 0, -6))
>>> job.comment
"Feed 'it'"
>>> job.check_trigger((2010, 5, 1, 7, 0), utc_offset=-6)
True
>>> job.check_trigger((2010, 5, 1, 16, 0), utc_offset=-6)
True
>>> job.check_trigger((2010, 5, 2, 1, 0), utc_offset=-6)
True
"""

import calendar
import datetime
import re

__all__ = ["CronExpression", "parse_atom", "DEFAULT_EPOCH", "SUBSTITUTIONS",
    "is_special_atom"]
__license__ = "MIT"

# Python 3 compatibility
if isinstance(map, type):
    def map(*args):
        return list(__builtins__['map'](*args))

    def zip(*args):
        return list(__builtins__['zip'](*args))

    xrange = range

DAY_NAMES = zip(('SUN', 'MON', 'TUE', 'WED', 'THU', 'FRI', 'SAT'), xrange(7))
MINUTES = (0, 59)
HOURS = (0, 23)
DAYS_OF_MONTH = (1, 31)
MONTHS = (1, 12)
DAYS_OF_WEEK = (0, 6)
L_FIELDS = (DAYS_OF_WEEK, DAYS_OF_MONTH)
FIELD_RANGES = (MINUTES, HOURS, DAYS_OF_MONTH, MONTHS, DAYS_OF_WEEK)
MONTH_NAMES = zip(('JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN',
                   'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC'), xrange(1, 13))
DEFAULT_EPOCH = (1970, 1, 1, 0, 0, 0)
SUBSTITUTIONS = {
    "@yearly": "0 0 1 1 *",
    "@annually": "0 0 1 1 *",
    "@monthly": "0 0 1 * *",
    "@weekly": "0 0 * * 0",
    "@daily": "0 0 * * *",
    "@midnight": "0 0 * * *",
    "@hourly": "0 * * * *"
}
VALIDATE_POUND = re.compile("^[0-6]#[1-5]")
VALIDATE_L_IN_DOW = re.compile("^[0-6]L$")
VALIDATE_W = re.compile("^[0-3]?[0-9]W$")


class CronExpression(object):
    def __init__(self, line, epoch=DEFAULT_EPOCH, epoch_utc_offset=0):
        """
        Instantiates a CronExpression object with an optionally defined epoch.
        If the epoch is defined, the UTC offset can be specified one of two
        ways: as the sixth element in 'epoch' or supplied in epoch_utc_offset.
        The epoch should be defined down to the minute sorted by
        descending significance.
        """
        for key, value in SUBSTITUTIONS.items():
            if line.startswith(key):
                line = line.replace(key, value)
                break

        fields = line.split(None, 5)
        if len(fields) == 5:
            fields.append('')

        minutes, hours, dom, months, dow, self.comment = fields

        dow = dow.replace('7', '0').replace('?', '*')
        dom = dom.replace('?', '*')

        for monthstr, monthnum in MONTH_NAMES:
            months = months.upper().replace(monthstr, str(monthnum))

        for dowstr, downum in DAY_NAMES:
            dow = dow.upper().replace(dowstr, str(downum))

        self.string_tab = [minutes, hours, dom, months, dow]
        self.compute_numtab()
        if len(epoch) == 5:
            y, mo, d, h, m = epoch
            self.epoch = (y, mo, d, h, m, epoch_utc_offset)
        else:
            self.epoch = epoch

    def __repr__(self):
        base = self.__class__.__name__ + "(%s)"
        cron_line = self.string_tab + [str(self.comment)]
        if not self.comment:
            cron_line.pop()
        arguments = '"' + ' '.join(cron_line) + '"'
        if self.epoch != DEFAULT_EPOCH:
            return base % (arguments + ", epoch=" + repr(self.epoch))
        else:
            return base % arguments

    def __str__(self):
        return repr(self)

    def compute_numtab(self):
        """
        Recomputes the sets for the static ranges of the trigger time.

        This method should only be called by the user if the string_tab
        member is modified.
        """
        self.numerical_tab = []

        for field_str, span in zip(self.string_tab, FIELD_RANGES):
            split_field_str = field_str.split(',')
            if len(split_field_str) > 1 and "*" in split_field_str:
                raise ValueError("\"*\" must be alone in a field.")

            unified = set()
            for cron_atom in split_field_str:
                # parse_atom only handles static cases
                if not(is_special_atom(cron_atom, span)):
                    unified.update(parse_atom(cron_atom, span))

            self.numerical_tab.append(unified)

        if self.string_tab[2] == "*" and self.string_tab[4] != "*":
            self.numerical_tab[2] = set()
        elif self.string_tab[4] == "*" and self.string_tab[2] != "*":
            self.numerical_tab[4] = set()

    def check_trigger(self, date_tuple, utc_offset=0):
        """
        Returns boolean indicating if the trigger is active at the given time.
        The date tuple should be in the local time. Unless periodicities are
        used, utc_offset does not need to be specified. If periodicities are
        used, specifically in the hour and minutes fields, it is crucial that
        the utc_offset is specified.
        """
        year, month, day, hour, mins = date_tuple
        given_date = datetime.date(year, month, day)
        zeroday = datetime.date(*self.epoch[:3])
        last_dom = calendar.monthrange(year, month)[-1]
        dom_matched = True

        # In calendar and datetime.date.weekday, Monday = 0
        given_dow = (datetime.date.weekday(given_date) + 1) % 7
        first_dow = (given_dow + 1 - day) % 7

        # Figure out how much time has passed from the epoch to the given date
        utc_diff = utc_offset - self.epoch[5]
        mod_delta_yrs = year - self.epoch[0]
        mod_delta_mon = month - self.epoch[1] + mod_delta_yrs * 12
        mod_delta_day = (given_date - zeroday).days
        mod_delta_hrs = hour - self.epoch[3] + mod_delta_day * 24 + utc_diff
        mod_delta_min = mins - self.epoch[4] + mod_delta_hrs * 60

        # Makes iterating through like components easier.
        quintuple = zip(
            (mins, hour, day, month, given_dow),
            self.numerical_tab,
            self.string_tab,
            (mod_delta_min, mod_delta_hrs, mod_delta_day, mod_delta_mon,
                mod_delta_day),
            FIELD_RANGES)

        for value, valid_values, field_str, delta_t, field_type in quintuple:
            # All valid, static values for the fields are stored in sets
            if value in valid_values:
                continue

            # The following for loop implements the logic for context
            # sensitive and epoch sensitive constraints. break statements,
            # which are executed when a match is found, lead to a continue
            # in the outer loop. If there are no matches found, the given date
            # does not match expression constraints, so the function returns
            # False as seen at the end of this for...else... construct.
            for cron_atom in field_str.split(','):
                if cron_atom[0] == '%':
                    if not(delta_t % int(cron_atom[1:])):
                        break

                elif '#' in cron_atom:
                    D, N = int(cron_atom[0]), int(cron_atom[2])
                    # Computes Nth occurence of D day of the week
                    if (((D - first_dow) % 7) + 1 + 7 * (N - 1)) == day:
                        break

                elif cron_atom[-1] == 'W':
                    target = min(int(cron_atom[:-1]), last_dom)
                    lands_on = (first_dow + target - 1) % 7
                    if lands_on == 0:
                        # Shift from Sun. to Mon. unless Mon. is next month
                        if target < last_dom:
                            target += 1
                        else:
                            target -= 2
                    elif lands_on == 6:
                        # Shift from Sat. to Fri. unless Fri. in prior month
                        if target > 1:
                            target -= 1
                        else:
                            target += 2

                    # Break if the day is correct, and target is a weekday
                    if target == day and (first_dow + target) % 7 > 1:
                        break

                elif cron_atom[-1] == 'L':
                    # In dom field, L means the last day of the month
                    target = last_dom

                    if field_type == DAYS_OF_WEEK:
                        # Calculates the last occurence of given day of week
                        desired_dow = int(cron_atom[:-1])
                        target = (((desired_dow - first_dow) % 7) + 29)
                        if target > last_dom:
                            target -= 7

                    if target == day:
                        break
            else:
                # See 2010.11.15 of CHANGELOG
                if field_type == DAYS_OF_MONTH and self.string_tab[4] != '*':
                    dom_matched = False
                    continue
                elif field_type == DAYS_OF_WEEK and self.string_tab[2] != '*':
                    # If we got here, then days of months validated so it does
                    # not matter that days of the week failed.
                    return dom_matched

                # None of the expressions matched which means this field fails
                return False

        # Arriving at this point means the date landed within the constraints
        # of all fields; the associated trigger should be fired.
        return True


def is_special_atom(cron_atom, span):
    """
    Returns a boolean indicating whether or not the string can be parsed by
    parse_atom to produce a static set. In the process of examining the
    string, the syntax of any special character uses is also checked.
    """
    for special_char in ('%', '#', 'L', 'W'):
        if special_char not in cron_atom:
            continue

        if special_char == '#':
            if span != DAYS_OF_WEEK:
                raise ValueError("\"#\" invalid where used.")
            elif not VALIDATE_POUND.match(cron_atom):
                raise ValueError("\"#\" syntax incorrect.")
        elif special_char == "W":
            if span != DAYS_OF_MONTH:
                raise ValueError("\"W\" syntax incorrect.")
            elif not(VALIDATE_W.match(cron_atom) and int(cron_atom[:-1]) > 0):
                raise ValueError("Invalid use of \"W\".")
        elif special_char == "L":
            if span not in L_FIELDS:
                raise ValueError("\"L\" invalid where used.")
            elif span == DAYS_OF_MONTH:
                if cron_atom != "L":
                    raise ValueError("\"L\" must be alone in days of month.")
            elif span == DAYS_OF_WEEK:
                if not VALIDATE_L_IN_DOW.match(cron_atom):
                    raise ValueError("\"L\" syntax incorrect.")
        elif special_char == "%":
            if not(cron_atom[1:].isdigit() and int(cron_atom[1:]) > 1):
                raise ValueError("\"%\" syntax incorrect.")
        return True
    else:
        return False



def parse_atom(parse, minmax):
    """
    Returns a set containing valid values for a given cron-style range of
    numbers. The 'minmax' arguments is a two element iterable containing the
    inclusive upper and lower limits of the expression.

    Examples:
    >>> parse_atom("1-5",(0,6))
    set([1, 2, 3, 4, 5])

    >>> parse_atom("*/6",(0,23))
    set([0, 6, 12, 18])

    >>> parse_atom("18-6/4",(0,23))
    set([18, 22, 0, 4])

    >>> parse_atom("*/9",(0,23))
    set([0, 9, 18])
    """
    parse = parse.strip()
    increment = 1
    if parse == '*':
        return set(xrange(minmax[0], minmax[1] + 1))
    elif parse.isdigit():
        # A single number still needs to be returned as a set
        value = int(parse)
        if value >= minmax[0] and value <= minmax[1]:
            return set((value,))
        else:
            raise ValueError("\"%s\" is not within valid range." % parse)
    elif '-' in parse or '/' in parse:
        divide = parse.split('/')
        subrange = divide[0]
        if len(divide) == 2:
            # Example: 1-3/5 or */7 increment should be 5 and 7 respectively
            increment = int(divide[1])

        if '-' in subrange:
            # Example: a-b
            prefix, suffix = [int(n) for n in subrange.split('-')]
            if prefix < minmax[0] or suffix > minmax[1]:
                raise ValueError("\"%s\" is not within valid range." % parse)
        elif subrange == '*':
            # Include all values with the given range
            prefix, suffix = minmax
        else:
            raise ValueError("Unrecognized symbol \"%s\"" % subrange)

        if prefix < suffix:
            # Example: 7-10
            return set(xrange(prefix, suffix + 1, increment))
        else:
            # Example: 12-4/2; (12, 12 + n, ..., 12 + m*n) U (n_0, ..., 4)
            noskips = list(xrange(prefix, minmax[1] + 1))
            noskips += list(xrange(minmax[0], suffix + 1))
            return set(noskips[::increment])
    else:
        raise ValueError("Atom \"%s\" not in a recognized format." % parse)



# -*- coding: utf-8 -*-
import unittest
import tempfile
import sys
from os import path

import shutil

from cpbox.tool import template
from cpbox.tool import file
from cpbox.tool import strings

from cpbox.tool.testutils import RandomDir

class TestTempalte(unittest.TestCase, RandomDir):

    def setUp(self):
        RandomDir.setUp(self)
        self.template_dir = path.join(self.tests_root_dir, 'test-data')

    def tearDown(self):
        RandomDir.tearDown(self)

    def test_write(self):
        import six
        target_file = path.join(self.temp_dir, self.random_str(10))
        strs = ['b', u'b', '好', u'好']
        for s in strs:
            file.file_put_contents(target_file, s)
            ret = file.file_get_contents(target_file)
            if six.PY2:
                self.assertEqual(type(ret), unicode)
            else:
                self.assertEqual(type(ret), str)

if __name__ == '__main__':
    unittest.main()



from cpbox.tool import functocli
from cpbox.tool import dockerutil
from cpbox.tool import template
from cpbox.app.env import EnvTool

redis_image = 'redis:3.0.1'


@functocli.keep_class
class RedisCli(object):

    def __init__(self, app, redis_config, port=6379, container_name='redis', data_dir=None):
        self.app = app
        self.redis_config = redis_config
        self.port = port
        self.redis_container = container_name
        self.data_dir = data_dir

    def _image_name(self):
        return EnvTool(self.app).docker_url() + redis_image

    def start_redis(self):
        config_in_container = '/usr/local/etc/redis/redis.conf'
        ports = [
            str(self.port) + ':6379',
        ]

        volumes = {
            self.redis_config: config_in_container,
        }

        if self.data_dir:
            volumes[self.data_dir] = '/data'

        args = dockerutil.base_docker_args(container_name=self.redis_container, ports=ports, volumes=volumes)
        cmd_data = {'image': self._image_name(), 'args': args, 'redis_config': config_in_container}
        cmd = template.render_str(
            'docker run -d --restart always {{ args }} {{ image }} redis-server {{ redis_config }}', cmd_data)
        self.app.shell_run(cmd)

    def send_cmd_to_redis_container(self, cmd):
        cmd = 'docker exec %s %s bash -c \'%s\'' % (dockerutil.fg_mode(), self.redis_container, cmd)
        self.app.shell_run(cmd)

    def to_redis(self):
        cmd = 'redis-cli'
        self.send_cmd_to_redis_container(cmd)

    def to_redis_cli(self):
        cmd = 'bash'
        self.send_cmd_to_redis_container(cmd)

    def stop_redis(self):
        self.app.stop_container(self.redis_container)
        self.app.remove_container(self.redis_container)

    def restart_redis(self):
        self.stop_redis()
        self.start_redis()



import socket
import os
import sys
import json

from os import path
from cpbox.tool import system
from cpbox.tool import net


def fg_mode():
    return '-it' if sys.stdout.isatty() else '-i'


def get_docker_network_gw(name):
    cmd = 'docker network inspect %s' % (name)
    code, ret = system.run_cmd(cmd)
    network_info = json.loads(ret)
    return network_info[0]['IPAM']['Config'][0]['Gateway']


def docker0_ip():
    if system.get_os_name() == 'mac':
        return get_docker_network_gw('bridge')
    else:
        return net.get_ip_address('docker0')


def base_docker_args(env=None, container_name=None, volumes=None, ports=None, current_user=False, envs=None,
                     working_dir=None, auto_hostname=True, hostname=None):
    host_hostname_fqdn = socket.gethostname()
    host_hostname_short = host_hostname_fqdn.split('.')[0]
    args = '--name ' + container_name
    if auto_hostname:
        args += ' -h %s.docker.%s' % (container_name, host_hostname_short)
    else:
        if hostname:
            args += ' -h %s' % (hostname)
        else:
            args += ' -h %s' % (host_hostname_short)

    args += ' -e PUPPY_HOSTNAME=' + host_hostname_fqdn
    args += ' -e PUPPY_CONTAINER_NAME=' + container_name
    for key, value in os.environ.items():
        if key.startswith('PUPPY_'):
            args += ' -e %s="%s"' % (key, value)
    if env:
        args += ' -e PUPPY_ENV=' + env

    if working_dir:
        args += ' -w ' + working_dir

    if current_user:
        args += ' -u %s:%s' % (os.getuid(), os.getgid())

    if volumes:
        if isinstance(volumes, dict):
            args += ' ' + ' '.join(('-v ' + from_ + ':' + to_ for from_, to_ in volumes.items()))
        else:
            args += ' ' + ' '.join(('-v ' + item for item in volumes))
    if ports:
        if isinstance(ports, dict):
            args += ' ' + ' '.join(('-p ' + str(from_) + ':' + str(to_) for from_, to_ in ports.items()))
        else:
            args += ' ' + ' '.join(('-p ' + str(item) for item in ports))
    if envs:
        args += ' ' + ' '.join(('-e ' + item for item in envs))

    os_name = system.get_os_name()
    if os_name != 'mac':
        args += ' -v /etc/localtime:/etc/localtime'
    if path.isfile('/etc/devops'):
        args += ' -v /etc/devops/:/etc/devops/:ro'

    return args



import argparse
import inspect
from ruamel import yaml
import os
import sys
import six

from distutils import util

support_types = ('str', 'int', 'bool', 'float')

from cpbox.tool import datatypes

def append_cli(container, cli):
    cli_map = {}
    if hasattr(container, '__cli_map__'):
        cli_map = container.__cli_map__
    for key, attr in cli.__class__.__dict__.items():
        if key[0:1] != '_' and callable(attr):
            cli_map[key] = cli
            setattr(container, key, attr.__get__(cli))

def merge_class(*proxy_cls_list):
    def real_decorator(cls):
        cls.__proxy_src_cls__ = cls
        cls.__proxy_cls_list__ = list(proxy_cls_list)
        return cls
    return real_decorator

def keep_class(cls):
    cls.keep_class_for_cli = True
    return cls

def keep_method(f):
    f.keep_method_for_cli = True
    return f

class AppSepc(object):

    def __init__(self, app, log_level, common_args_option=None):

        self.common_args_option = common_args_option

        self.app = app
        self.get_app_public_methods(app)
        self.method_spec_map = self._get_app_method_spec()
        self.num_to_name_map = {str(index): name for index, name in enumerate(self.public_methods)}

        self.method_will_call = None

    def get_app_public_methods_for_proxy_class(self, app):
        proxy_src_cls = app.__dict__['__proxy_src_cls__']
        proxy_cls_list = app.__dict__['__proxy_cls_list__']

        method_names = []
        method_map = {}

        src_cls_methods = get_cls_cli_methods(proxy_src_cls, True)
        method_names.extend(src_cls_methods.keys())
        method_map.update(src_cls_methods)

        for cls in proxy_cls_list:
            cls_methods = get_cls_cli_methods(cls, False)
            method_names.extend(cls_methods.keys())
            method_map.update(cls_methods)

        self.public_methods = method_names
        self.method_map = method_map

    def get_app_public_methods_for_simple_class(self, app):
        cls_methods = get_cls_cli_methods(app, True)
        self.public_methods = cls_methods.keys()
        self.method_map = cls_methods

    def get_app_public_methods(self, app):
        if '__proxy_cls_list__' in app.__dict__:
            self.get_app_public_methods_for_proxy_class(app)
        else:
            self.get_app_public_methods_for_simple_class(app)

    def find_method_name(self, num_or_name):
        if num_or_name in self.num_to_name_map:
            return self.num_to_name_map[num_or_name]
        if num_or_name in self.method_spec_map:
            return num_or_name
        return None

    def fix_terminal_columns(self):
        if sys.stdin.isatty() and sys.stdout.isatty():
            rows, columns = os.popen('stty size', 'r').read().split()
            os.environ['COLUMNS'] = columns

    def resolve(self, default_method=None, enable_method_index=False):
        self.fix_terminal_columns()

        method_will_call = None
        if len(sys.argv) > 1:
            method_will_call = sys.argv[1].lstrip('-')
            method_will_call = method_will_call.replace('-', '_')
        if method_will_call is None and default_method is not None:
            method_will_call =  default_method
        method_will_call = self.find_method_name(method_will_call)

        if not method_will_call:
            self.show_help_then_exit(enable_method_index)

        self.method_will_call = method_will_call
        self._parse_args()

    def _parse_args(self):
        method_will_call = self.method_will_call
        method_spec = self.method_spec_map[method_will_call]
        kwargs = build_argparser_for_func(method_will_call, method_spec)

        self.common_kwargs = {}
        if self.common_args_option:
            common_args_keys = set(self.common_args_option.get('args'))
            methods_args_keys = set(method_spec['method_args'])
            method_kwargs = {}
            for key, value in kwargs.items():
                if key in common_args_keys:
                    self.common_kwargs[key] = value
                if key in methods_args_keys:
                    method_kwargs[key] = value
            self.kwargs = method_kwargs
        else:
            self.kwargs = kwargs

    def run(self):
        kwargs = self.kwargs
        _app = self.app(**self.common_kwargs)
        func = getattr(_app, self.method_will_call)
        func(**kwargs)

    def _get_app_method_spec(self):
        method_spec_map = {}
        for method_name in self.public_methods:
            method_spec_map[method_name] = self.get_method_spec(method_name)
        return method_spec_map

    def get_method_spec(self, method_name):
        method_attr = self.method_map[method_name]
        method_arg_spec = inspect.getargspec(method_attr)
        default_values = None
        if method_arg_spec.defaults:
            default_values = dict(zip(method_arg_spec.args[-len(method_arg_spec.defaults):], method_arg_spec.defaults))
        args = list(filter(lambda x: x != 'self', method_arg_spec.args))

        method_spec = {}
        method_spec['method_args'] = list(args)
        if self.common_args_option != None:
            args.extend(self.common_args_option.get('args', []))
            args = list(set(args))
            default_values_from_common_args = self.common_args_option.get('default_values', [])
            if default_values is None:
                default_values = default_values_from_common_args
            else:
                default_values.update(default_values_from_common_args)

        method_spec['args'] = args
        method_spec['default_values'] = default_values
        method_spec['parameters'] = _check_parameters_spec(method_name, args, default_values)
        return method_spec

    def show_help_then_exit(self, enable_method_index):
        parser = argparse.ArgumentParser(add_help=False)
        parser.allow_abbrev = False
        parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
        for index, method_name in enumerate(self.public_methods):
            if enable_method_index:
                parser.add_argument('-' + str(index), '--' + method_name, action='store_true')
            else:
                parser.add_argument('--' + method_name, action='store_true')
        parser.print_help()

        method_will_call = self.method_will_call
        if method_will_call and (method_will_call == 'h' or method_will_call == 'help'):
            sys.exit(0)
        else:
            sys.exit(1)

def _get_func_comments(method_attr):
    comments = inspect.getdoc(method_attr)
    if comments is None:
        return None
    else:
        return yaml.load(comments)

def _is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def _check_parameter_item_default_value(arg_name, parameter_item, default_values):
    if default_values is None or arg_name not in default_values:
        return parameter_item

    value = default_values[arg_name]
    type = _determine_type(value)

    if type is None:
        raise Error('Can not determine type for parameter: %s, parameter_item: %s' %s (arg_name, parameter_item))

    parameter_item['type'] = type
    parameter_item['des'] = 'type: %s, default value is: %s' % (type, value)
    return parameter_item

def _determine_type(value):
    type = None
    if value is None:
        type = 'str'
    elif isinstance(value, bool):
        type = 'bool'
    elif isinstance(value, six.integer_types):
        type = 'int'
    elif isinstance(value, six.string_types):
        type = 'str'
    elif _is_float(value):
        type = 'float'
    return type

def _convert_value_for_type(value, type):
    if value is None:
        return value
    elif type == 'bool':
        if isinstance(value, six.string_types):
            try:
                return bool(util.strtobool(value))
            except:
                return False
        return bool(value)
    elif type == 'int':
        return int(value)
    elif type == 'str':
        return str(value)
    elif type == 'float':
        return float(value)
    return value

def _check_parameters_spec(method_name, args, default_values):
    parameters = {}
    for arg_name in args:
        parameter_item = {}
        if isinstance(parameter_item, six.string_types):
            parameter_item = {
                    'type': 'str',
                    'des': parameter_item
                    }
        parameter_item = _check_parameter_item_default_value(arg_name, parameter_item, default_values)
        if 'type' not in parameter_item:
            parameter_item['type'] = 'str'
        if 'des' not in parameter_item:
            parameter_item['des'] = arg_name

        parameters[arg_name] = parameter_item
        _check_parameter_item(arg_name, parameter_item)

    return parameters

def _check_parameter_item(arg_name, parameter_item):
    if 'type' not in parameter_item:
        print('Missing type in parameter item: %s' % parameter_item)
    if parameter_item['type'] not in support_types:
        print('The type: %s is not in support types: %s' % (parameter_item['type'], support_types))
    if not parameter_item['des']:
        print('Missing parameter descripion(`des`): %s' % parameter_item)

def get_cls_cli_methods(cls, exclude_parent=True):
    methods = get_cls_public_methods(cls)
    if exclude_parent:
        parenet_cls_methods = get_cls_parent_methods_for_exclude(cls)
        return {key: value for key, value in methods.items() if key not in parenet_cls_methods}
    else:
        return methods

def get_cls_public_methods(cls):
    list = inspect.getmembers(cls, predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x))
    methods = {item[0]: item[1] for item in list if item[0][0:1] != '_'}
    return methods

def get_cls_parent_methods_for_exclude(cls):
    methods = {}
    for parent_cls in cls.__bases__:
        if parent_cls.__dict__.get('keep_class_for_cli', False):
            continue
        parent_methods = get_cls_public_methods(parent_cls)
        parent_methods = {key: val for key, val in parent_methods.items() if not val.__dict__.get('keep_method_for_cli', False)}
        methods.update(parent_methods)
    return methods

def build_argparser_for_func(method_name, method_spec):

    args = method_spec['args']
    default_values = method_spec['default_values']

    parser = argparse.ArgumentParser()
    parser.add_argument('--' + method_name, action='store_true')

    parameters = method_spec.get('parameters', {})
    for arg_name in args:
        parameter = parameters.get(arg_name)
        help = parameter.get('des', '')
        metavar = ''

        if default_values is not None and arg_name in default_values:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
        else:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
    args_input, _ = parser.parse_known_args()
    kwargs = {}
    for arg_name in args:
        parameter = parameters.get(arg_name)
        attr = getattr(args_input, arg_name)
        if default_values is not None and arg_name in default_values and attr == default_values[arg_name]:
            kwargs[arg_name] = attr
        else:
            kwargs[arg_name] = _convert_value_for_type(
                    _get_first_from_list_or_item_self(attr), parameter['type'])
    return kwargs

def _get_first_from_list_or_item_self(attr):
    try:
        return next(iter(attr), None)
    except TypeError:
        return attr

def _check_args_has(args, option):
    ret = getattr(args, option)
    return ret

def run_app(app, log_level='critical', default_method=None, common_args_option=None, common_args_option_basic=None):
    common_args_option = common_args_option if common_args_option else {}
    if common_args_option_basic:
        common_args_option = datatypes.merge(common_args_option_basic, common_args_option, 'merge')
    app_spec = AppSepc(app, log_level, common_args_option)
    app_spec.resolve(default_method)
    app_spec.run()


import socket
import fcntl
import struct

def get_ip_address(ifname):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', ifname.encode('utf-8')[:15])
        )[20:24])
    except IOError as e:
        raise e
    finally:
        s.close()
    return None


def get_hostname_ip():
    ip = socket.gethostbyname(socket.gethostname())
    return ip

def get_local_ip_address():
    try:
        return get_ip_address_udp()
    finally:
        return '127.0.0.1'

def get_ip_address_udp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    ret = s.getsockname()[0]
    s.close()
    return ret


def is_open(ip, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        return True
    except:
        pass
    finally:
        s.close()
    return False



import json
import redis

class Proxy():

    def __init__(self):
        self.redis = None

    def init_redis(self, from_url):
        pool = redis.ConnectionPool.from_url(from_url)
        self.redis = redis.Redis(connection_pool=pool)

proxy = Proxy()

def set_proxy(redis):
    proxy.redis = redis

def init_proxy(from_url):
    proxy.init_redis(from_url)

def flushall():
    # not use in prod
    proxy.redis.flushall()

def set(kv, value, ttl = 600):
    value = json.dumps(value)
    proxy.redis.set(kv, value, ttl)

def set_list(kv_list, ttl = 600):
    pipe = proxy.redis.pipeline()
    for (k, value) in kv_list.items():
        value = json.dumps(value)
        pipe.set(k, value, ttl)
    pipe.execute()

def delete_keys(keys):
    pipe = proxy.redis.pipeline()
    for key in keys:
        pipe.delete(key)
    pipe.execute()

def delete(k):
    proxy.redis.delete(k)

def get(key, fallback=None):
    value = proxy.redis.get(key)
    if value is None:
        return fallback
    return json.loads(value)

def incr(key, value = 1, ttl = 600):
    proxy.redis.incr(key, value)
    proxy.redis.expire(key, ttl)

def get_list(key_list):
    redis_value_list = proxy.redis.mget(key_list)
    return_value_list = {}
    for index, key in enumerate(key_list):
        if redis_value_list[index] is None:
            continue
        else:
            return_value_list[key] = json.loads(redis_value_list[index])
    return return_value_list

def delete_list(k_list):
    pipe = proxy.redis.pipeline()
    for k in k_list:
        pipe.delete(k)
    pipe.execute()



# -*- coding: utf-8 -*-
import unittest
from os import path

from cpbox.tool import template
from cpbox.tool.testutils import RandomDir


class TestTempalte(unittest.TestCase, RandomDir):

    def setUp(self):
        RandomDir.setUp(self)
        self.template_dir = path.join(self.tests_root_dir, 'test-data')

    def tearDown(self):
        RandomDir.tearDown(self)

    def test_render_to_file(self):
        data = {'date': '20190520'}
        str_after_render = u'寒雨连江夜入吴 平明送客楚山孤 20190520'

        template_file = path.join(self.template_dir, 'test-template-unicode.jinja2')
        ret = template.render_to_str(template_file, data)
        self.assertEqual(ret, str_after_render)

        target_file = path.join(self.temp_dir, 'tmp')
        ret_from_file = template.render_to_file(template_file, data, target_file)
        self.assertEqual(ret_from_file, str_after_render)

if __name__ == '__main__':
    unittest.main()
